#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 23 16:23:35 2018

@author: anazabal, olmosUC3M, ivaleraM
"""

# # -*- coding: utf-8 -*-

## 2 fully connected layers at both encoding and decoding
# hidden_dim is the number of neurons of the first hidden layer

import tensorflow as tf
from hivae import VAE_functions
        
def encoder(X_list, miss_list, batch_size, z_dim, s_dim, tau):
    
    samples = dict.fromkeys(['s','z','y','x'],[])
    q_params = dict()
    X = tf.concat(X_list,1)
    
    #Create the proposal of q(s|x^o)
    samples['s'], q_params['s'] = VAE_functions.s_proposal_multinomial(X, batch_size, s_dim, tau, reuse=None)
    
    #Create the proposal of q(z|s,x^o)
    samples['z'], q_params['z'] = VAE_functions.z_proposal_GMM_factorized(X_list, samples['s'], miss_list, batch_size, z_dim, reuse=None)
    
    return samples, q_params
        

def decoder(batch_data_list, miss_list, types_list, samples, q_params, normalization_params, batch_size, z_dim, y_dim, y_dim_partition, tau2):
    
    p_params = dict()
    
    #Create the distribution of p(z|s)
    p_params['z'] = VAE_functions.z_distribution_GMM(samples['s'], z_dim, reuse=None)
    
    #Create deterministic layer y
    samples['y'] = tf.layers.dense(inputs=samples['z'], units=y_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name= 'layer_h1_', reuse=None)
    
    grouped_samples_y = VAE_functions.y_partition(samples['y'], types_list, y_dim_partition)

    #Compute the parameters h_y
#    theta = VAE_functions.theta_estimation_from_y(grouped_samples_y, types_list, miss_list, batch_size, reuse=None)
    theta = VAE_functions.theta_estimation_from_ys(grouped_samples_y, samples['s'], types_list, miss_list, batch_size, reuse=None)
    
    #Compute loglik and output of the VAE
    log_p_x, log_p_x_missing, samples['x'], p_params['x'] = VAE_functions.loglik_evaluation(batch_data_list, types_list, miss_list, theta, tau2, normalization_params, reuse=None)
        
    return theta, samples, p_params, log_p_x, log_p_x_missing

def cost_function(log_p_x, p_params, q_params, types_list, z_dim, y_dim, s_dim):
    
    #KL(q(s|x)|p(s))
    log_pi = q_params['s']
    pi_param = tf.nn.softmax(log_pi)
    KL_s = -tf.nn.softmax_cross_entropy_with_logits(logits=log_pi, labels=pi_param) + tf.log(float(s_dim))
    
    #KL(q(z|s,x)|p(z|s))
    mean_pz, log_var_pz = p_params['z']
    mean_qz, log_var_qz = q_params['z']
    KL_z = -0.5*z_dim +0.5*tf.reduce_sum(tf.exp(log_var_qz - log_var_pz) +tf.square(mean_pz - mean_qz)/tf.exp(log_var_pz) -log_var_qz + log_var_pz,1)
    
    #Eq[log_p(x|y)]
    loss_reconstruction = tf.reduce_sum(log_p_x,0)
    
    #Complete ELBO
    ELBO = tf.reduce_mean(loss_reconstruction - KL_z - KL_s,0)
    
    return ELBO, loss_reconstruction, KL_z, KL_s

def samples_generator(batch_data_list, X_list, miss_list, types_list, batch_size, z_dim, y_dim, y_dim_partition, s_dim, tau, tau2, normalization_params):
    
    samples_test = dict.fromkeys(['s','z','y','x'],[])
    test_params = dict()
    X = tf.concat(X_list,1)
    
    #Create the proposal of q(s|x^o)
    _, params = VAE_functions.s_proposal_multinomial(X, batch_size, s_dim, tau, reuse=True)
    samples_test['s'] = tf.one_hot(tf.argmax(params,1),depth=s_dim)
    
    #Create the proposal of q(z|s,x^o)
    _, params = VAE_functions.z_proposal_GMM_factorized(X_list, samples_test['s'], miss_list, batch_size, z_dim, reuse=True)
    samples_test['z'] = params[0]
    
    #Create deterministic layer y
    samples_test['y'] = tf.layers.dense(inputs=samples_test['z'], units=y_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), trainable=True, name= 'layer_h1_', reuse=True)
    
    grouped_samples_y = VAE_functions.y_partition(samples_test['y'], types_list, y_dim_partition)
    
    #Compute the parameters h_y
#    theta = VAE_functions.theta_estimation_from_y(grouped_samples_y, types_list, miss_list, batch_size, reuse=True)
    theta = VAE_functions.theta_estimation_from_ys(grouped_samples_y, samples_test['s'], types_list, miss_list, batch_size, reuse=True)
    
    
    #Compute loglik and output of the VAE
    log_p_x, log_p_x_missing, samples_test['x'], test_params['x'] = VAE_functions.loglik_evaluation(batch_data_list, types_list, miss_list, theta, tau2, normalization_params, reuse=True)
    
    return samples_test, test_params, log_p_x, log_p_x_missing



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  5 16:35:35 2017

Grouping all parsers specifications here

@author: anazabal, olmosUC3M, ivaleraM
"""

import argparse

def getArgs(argv=None):
    parser = argparse.ArgumentParser(description='Default parameters of the models',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--batch_size', type=int, default=200, help='Size of the batches')
    parser.add_argument('--epochs',type=int,default=5001, help='Number of epochs of the simulations')
    parser.add_argument('--perp',type=int,default=10, help='Perplexity for the t-SNE')
    parser.add_argument('--train', type=int,default=1, help='Training model flag')
    parser.add_argument('--display', type=int,default=1, help='Display option flag')
    parser.add_argument('--save', type=int,default=1000, help='Save variables every save iterations')
    parser.add_argument('--restore', type=int,default=0, help='To restore session, to keep training or evaluation') 
    parser.add_argument('--plot', type=int,default=1, help='Plot results flag')
    parser.add_argument('--dim_latent_s',type=int,default=10, help='Dimension of the categorical space')
    parser.add_argument('--dim_latent_z',type=int,default=2, help='Dimension of the Z latent space')
    parser.add_argument('--dim_latent_y',type=int,default=10, help='Dimension of the Y latent space')
    parser.add_argument('--dim_latent_y_partition',type=int, nargs='+', help='Partition of the Y latent space')
    parser.add_argument('--miss_percentage_train',type=float,default=0.0, help='Percentage of missing data in training')
    parser.add_argument('--miss_percentage_test',type=float,default=0.0, help='Percentage of missing data in test')
    parser.add_argument('--model_name', type=str, default='model_new', help='File of the training model')
    parser.add_argument('--save_file', type=str, default='new_mnist_zdim5_ydim10_4images_', help='Save file name')
    parser.add_argument('--data_file', type=str, default='MNIST_data', help='File with the data')
    parser.add_argument('--types_file', type=str, default='mnist_train_types2.csv', help='File with the types of the data')
    parser.add_argument('--miss_file', type=str, default='Missing_test.csv', help='File with the missing indexes mask')
    parser.add_argument('--true_miss_file', type=str, help='File with the missing indexes when there are NaN in the data')
    parser.add_argument('--result_dir', type=str, default="./results_default/", help='Path for results')
    parser.add_argument('--latent_type', type=str, default=None, help='Output latent variables - otherwise a string is expected, ie. train or test, for differentiating between the embedding in trhe different datasets')

    return parser.parse_args(argv)



def close_all():
    """Try to flush and close all file handles"""
    global FILEHANDLES
    for fh in FILEHANDLES:
        try:
            fh.flush()
        except Exception:
            pass
        try:
            fh.close()
        except Exception:
            pass


from .hivae import save_get, hivae
# import graph_new
# import VAE_functions
# import model_HIVAE_factorized
# import model_HIVAE_inputDropout
# import loglik_models_missing_normalize
# import read_functions
# import utils




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""New library for HI-VAE

@author: fathyshalaby,athro
"""
import sys
# import os
# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3' 
import tensorflow as tf
#tf.logging.set_verbosity(tf.logging.ERROR)
# import pprint
# pprinter = pprint.PrettyPrinter(depth=3)
from hivae import graph_new
import time
import numpy as np
import pandas as pd
#import plot_functions
from hivae import read_functions
import os
import csv
from pathlib import Path
import uuid



#ak: small method to savely get information from a dictionary
def save_get(aHash,aKey,aDefault):
    try:
        return aHash[aKey]
    except:
        pass
    return aDefault

class hivae():
    def __init__(self,
                 types_description,
                 network_dict,
                 results_path     = None,          #ak: included a path for results
                 save_every_epoch = 1000,
                 network_path     = None,          # may be internal?
                 verbosity_level  = 1
                 ):

        self.model_name       = save_get(network_dict,'model_name','unkown_model_name')
        self.m_perc           = 0
        self.mask             = save_get(network_dict,'mask',None)
        self.types_list       = read_functions.get_types_list(types_description)
        self.save             = 1000 # parameter?
        self.save_every_epoch = save_every_epoch    
        self.dim_s            = save_get(network_dict,'dim_s',None) 
        self.dim_z            = save_get(network_dict,'dim_z',None) 
        self.dim_y            = save_get(network_dict,'dim_y',None)
        self.batch_size       = save_get(network_dict,'batch_size',32)
        self.verbosity_level  = verbosity_level


        # self.true_miss_file   = true_miss_file
        # self.miss_file        = miss_file
        if not results_path:
            results_path     = './hivae_results'          #ak: included a path for results
        if not network_path:
            network_path     = './hivae_networks'

                         

        
        self.network_path     = network_path
        self.results_path     = results_path
        self.display          = True
        
        ###ak: not sure why the training method is called while constructing the string
        #self.savefile = str(str(self.model_name)+'_'+'_Missing'+str(self.m_perc)+'_'+str(self.mask)+'_z'+str(self.dim_z)+'_y'+str(self.dim_y)+'_s'+str(self.dim_s)+'_batch'+str(self.training()[1]))
        self.savefile               = str(str(self.model_name)+'_'+'_Missing'+str(self.m_perc)+'_'+str(self.mask)+'_z'+str(self.dim_z)+'_y'+str(self.dim_y)+'_s'+str(self.dim_s)+'_batch'+str(save_get(network_dict,'batch_size','batch_size_unkown')))
        self.experiment_name_plain  = '{}_s{}_z{}_y{}_batch{}'.format(self.model_name,self.dim_s,self.dim_z,self.dim_y,self.batch_size)
        # generate random UUID
        some_hash                   = uuid.uuid4().urn
        #get string representation - removing leading 'urn:uuid:'
        remove_urn_string = 'urn:uuid:'
        some_hash = some_hash[some_hash.rfind(remove_urn_string)+len(remove_urn_string):]
        
        self.experiment_name_hash   = '{}_s{}_z{}_y{}_batch{}_{}'.format(self.model_name,self.dim_s,self.dim_z,self.dim_y,self.batch_size,some_hash)
        self.experiment_name        = self.experiment_name_hash
        
        # Create a directoy for the network files
        full_network_path = '{}/{}'.format(self.network_path,self.experiment_name)
        try:
            if not os.path.exists(full_network_path):
                os.makedirs(full_network_path)
        except:
            self.vprint(3,'Could not create network path <<{}>>'.format(full_network_path))
            pass
        
        # Create a directoy for the results_path 
        full_results_path = '{}/{}'.format(self.results_path,self.experiment_name)
        try:
            if not os.path.exists(full_results_path):
                os.makedirs(full_results_path)
        except:
            self.vprint(3,'Could not create results path <<{}>>'.format(full_results_path))
            pass
        
        self.full_network_path = full_network_path
        self.full_results_path = full_results_path

        self.network_file_name = '{}_ckpt'.format(self.full_network_path) 
        self.log_file_name     = '{}/log_file_{}.txt'.format(self.full_network_path,self.savefile)

        print('self.full_network_path',self.full_network_path)
        print('self.full_results_path',self.full_results_path)
        print('self.network_file_name', self.network_file_name)


    def set_verbosity_level(self,verbosity_level=1):
        # if self.verbosity_level>= 0 and self.verbosity_level<=5:
        self.verbosity_level = verbosity_level
        
    def vprint(self,verbosity_level,*args):
        if verbosity_level<=self.verbosity_level:
            if verbosity_level==1:
                print('INFO :\t',*args)
            elif verbosity_level==2:
                print('DEBUG:\t',*args)
            elif verbosity_level==3:
                print('ERROR:\t',*args)


            

        
    def print_loss(self,epoch, start_time, avg_loss, avg_test_loglik, avg_KL_s, avg_KL_z):
        
        self.vprint(2,'Epoch: {:4d}\ttime: {:4.2f}\ttrain_loglik: {:5.2f}\tKL_z: {:5.2f}\tKL_s: {:5.2f}\tELBO: {:5.2f}\tTest_loglik: {:5.2f}'.format(epoch, time.time() - start_time, avg_loss, avg_KL_z, avg_KL_s, avg_loss-avg_KL_z-avg_KL_s, avg_test_loglik))


    #ak: save the data in a more coherent fashion
    def save_data(self,the_main_directory,the_path,the_data):
        self.vprint(1,'Saving {:40s} in {:40s}'.format(the_path,the_main_directory))
        dir_present = False
        try:
            if not os.path.exists(the_main_directory):
                os.makedirs(the_main_directory)
            dir_present = True
        except:
            self.vprint(3,'Could not create path (<<{}>>)'.format(the_main_directory))

        if dir_present:
            full_file_path = '{}/{}'.format(the_main_directory,the_path)
            try:
                with open(full_file_path, 'w') as f:
                    writer = csv.writer(f)
                    writer.writerows(the_data)
            except:
                self.vprint(3,'Problens writing data to <<{}>>\n{}'.format(full_file_path,str(the_data)))
            
            

    def fit(self,
                training_data,
                epochs            = 200,
                learning_rate     = 1e-3,
                # results_path      = './hivae_networks/',
                # restore_session   = False,
                true_missing_mask = None,
                missing_mask      = None,
                verbosity         = 0):

        return self._training(training_data,
                    epochs            = epochs,
                    learning_rate     = learning_rate,
                    results_path      = self.full_results_path,
                    restore           = False,
                    train_or_test     = True,
                    restore_session   = False,
                    true_missing_mask = true_missing_mask,
                    missing_mask      = missing_mask,
                    verbosity         = verbosity)

    def predict(self,
                testing_data,
                # epochs            = 1,
                learning_rate     = 1e-3,
                # results_path      = './hivae_networks/',
                # restore_session   = True,
                true_missing_mask = None,
                missing_mask      = None,
                verbosity         = 0):

        
        return self._training(testing_data,
                    epochs            = 1,
                    learning_rate     = learning_rate,
                    results_path      = self.full_results_path,
                    restore           = True,
                    train_or_test     = False,
                    restore_session   = True,
                    true_missing_mask = true_missing_mask,
                    missing_mask      = missing_mask,
                    verbosity         = verbosity)
        
                

    def _training(self,
                      training_data,epochs=200,
                      learning_rate=1e-3,
                      results_path='./',
                      restore=False,
                      train_or_test=True,
                      restore_session=False,
                      true_missing_mask=None,
                      missing_mask = None,
                      verbosity=3):
        
        self.vprint(2,'len(training_data)',len(training_data))

        # train_or_test == True - training
        training_phase  = train_or_test
        testing_phase   = not train_or_test

        batch_size_here = self.batch_size
        
        # if testing phase use all as one single batch
        if testing_phase:
            batch_size_here = len(training_data)
        
        # train_data = None
        # # test whether data is pandas DataFrame or numpy.nparray
        # if type(training_data) == type(pd.DataFrame()):
        #     train_data = training_data.to_numpy()
        # elif type(training_data) == type((np.ndarray([1]))):
        #     train_data = training_data
        # else:
        #     print('DEBUG-AK: Data not in correct format, Type = {}'.format(type(training_data)))
        #     sys.exit(-100)

        # train_data, _ , miss_mask, true_miss_mask, n_samples = read_functions.read_data_df_as_input(training_data,
        #                                                                                         self.types_list,
        #                                                                                         self.miss_file,
        #                                                                                         self.true_miss_file)

        true_miss_set = False
        miss_set      = False
        # deal with true missing data
        if type(true_missing_mask)==type(pd.DataFrame()) and not true_missing_mask.empty:
            true_miss_mask = true_missing_mask.to_numpy()
            true_miss_set = True
        else:
            true_miss_mask = np.ones(training_data.shape,dtype=int)

        # deal with non-true missing data
        if type(missing_mask)==type(pd.DataFrame()) and not missing_mask.empty:
            miss_mask = missing_mask.to_numpy()
            miss_set = True
        else:
            miss_mask = np.ones(training_data.shape,dtype=int)

        train_data, _ , n_samples = read_functions.read_data_df(training_data,self.types_list,true_miss_mask)

            
        n_batches = int(np.floor(np.shape(train_data)[0]/batch_size_here))
        #Compute the real miss_mask
        miss_mask = np.multiply(miss_mask, true_miss_mask)
        #Creating graph
        sess_HVAE = tf.Graph()
        
        with sess_HVAE.as_default():

            tf_nodes = graph_new.HVAE_graph(
                self.model_name,
                self.types_list,
                batch_size_here,
                learning_rate=learning_rate,
                z_dim=self.dim_z,
                y_dim=self.dim_y,
                s_dim=self.dim_s,
                y_dim_partition=None)

        ################### Running the VAE Training #################################
        with tf.Session(graph=sess_HVAE) as session:
        
            # Add ops to save and restore all the variables.
            saver = tf.train.Saver()
        
            if training_phase:
                self.vprint(1,'Training the HVAE ...')
            elif testing_phase:
                self.vprint(1,'Testing the HVAE ...')
                
            if restore_session:
                saver.restore(session, self.network_file_name)
                self.vprint(1,"Model restored.")
            else:
                self.vprint(1,'Initizalizing Variables ...')
                tf.global_variables_initializer().run()
    
        
            start_time = time.time()
            # Training cycle
        
            loglik_epoch = []
            testloglik_epoch = []
            error_train_mode_global = []
            error_test_mode_global = []
            KL_s_epoch = []
            KL_z_epoch = []

            for epoch in range(epochs):
                avg_loss = 0.
                avg_KL_s = 0.
                avg_KL_y = 0.
                avg_KL_z = 0.
                samples_list = []
                p_params_list = []
                q_params_list = []
                log_p_x_total = []
                log_p_x_missing_total = []
            
                # Annealing of Gumbel-Softmax parameter
                tau = None
                if training_phase:
                    tau = np.max([1.0 - 0.01*epoch,1e-3])
                elif testing_phase:
                    tau = 1e-3
                #            tau = 1e-3
                tau2 = np.min([0.001*epoch,1.0])

            
                #Randomize the data in the mini-batches
                random_perm = np.random.permutation(range(np.shape(train_data)[0]))
                train_data_aux = train_data[random_perm,:]
                miss_mask_aux = miss_mask[random_perm,:]
                true_miss_mask_aux = true_miss_mask[random_perm,:]
            
                for i in range(n_batches):      
                
                    #Create inputs for the feed_dict
                    data_list, miss_list = read_functions.next_batch(train_data_aux, self.types_list, miss_mask_aux, batch_size_here, index_batch=i)

                    #Delete not known data (input zeros)
                    data_list_observed = [data_list[i]*np.reshape(miss_list[:,i],[batch_size_here,1]) for i in range(len(data_list))]
                
                    #Create feed dictionary
                    feedDict = {i: d for i, d in zip(tf_nodes['ground_batch'], data_list)}
                    feedDict.update({i: d for i, d in zip(tf_nodes['ground_batch_observed'], data_list_observed)})
                    feedDict[tf_nodes['miss_list']] = miss_list
                    feedDict[tf_nodes['tau_GS']] = tau
                    feedDict[tf_nodes['tau_var']] = tau2
                    
                    #Running VAE
                    loss,KL_z,KL_s,samples,log_p_x,log_p_x_missing,p_params,q_params = None,None,None,None,None,None,None,None
                    if training_phase:
                        _,loss,KL_z,KL_s,samples,log_p_x,log_p_x_missing,p_params,q_params  = session.run([tf_nodes['optim'],
                                                                                                               tf_nodes['loss_re'], tf_nodes['KL_z'],
                                                                                                               tf_nodes['KL_s'], tf_nodes['samples'],
                                                                                                               tf_nodes['log_p_x'], tf_nodes['log_p_x_missing'],
                                                                                                               tf_nodes['p_params'],tf_nodes['q_params']],
                                                                                                              feed_dict=feedDict)
                    elif testing_phase:
                        loss,KL_z,KL_s,samples,log_p_x,log_p_x_missing,p_params,q_params  = session.run([tf_nodes['loss_re'], tf_nodes['KL_z'],
                                                                                                             tf_nodes['KL_s'], tf_nodes['samples'],
                                                                                                             tf_nodes['log_p_x'], tf_nodes['log_p_x_missing'],
                                                                                                             tf_nodes['p_params'],tf_nodes['q_params']],
                                                                                                            feed_dict=feedDict)


                    samples_test,log_p_x_test,log_p_x_missing_test,test_params  = session.run([tf_nodes['samples_test'],
                                                                                                   tf_nodes['log_p_x_test'],tf_nodes['log_p_x_missing_test'],
                                                                                                   tf_nodes['test_params']],
                                                                                                  feed_dict=feedDict)
                
                    #Evaluate results on the imputation with mode, not on the samlpes!
                    samples_list.append(samples_test)
                    p_params_list.append(test_params)
                    #                        p_params_list.append(p_params)
                    q_params_list.append(q_params)
                    log_p_x_total.append(log_p_x_test)
                    log_p_x_missing_total.append(log_p_x_missing_test)
                
                    # Compute average loss
                    avg_loss += np.mean(loss)
                    avg_KL_s += np.mean(KL_s)
                    avg_KL_z += np.mean(KL_z)
                
                #Concatenate samples in arrays
                s_total, z_total, y_total, est_data = read_functions.samples_concatenation(samples_list)
                #Transform discrete variables back to the original values
                train_data_transformed = read_functions.discrete_variables_transformation(train_data_aux[:n_batches*batch_size_here,:], self.types_list)
                est_data_transformed = read_functions.discrete_variables_transformation(est_data, self.types_list)
                est_data_imputed = read_functions.mean_imputation(train_data_transformed, miss_mask_aux[:n_batches*batch_size_here,:], self.types_list)
            
                #est_data_transformed[np.isinf(est_data_transformed)] = 1e20
            
                #Create global dictionary of the distribution parameters
                p_params_complete = read_functions.p_distribution_params_concatenation(p_params_list, self.types_list, self.dim_z, self.dim_s)
                q_params_complete = read_functions.q_distribution_params_concatenation(q_params_list,  self.dim_z, self.dim_s)
            
                #Number of clusters created
                cluster_index = np.argmax(q_params_complete['s'],1)
                cluster = np.unique(cluster_index)
                self.vprint(2,'Clusters: ' + str(len(cluster)))
            
                #Compute mean and mode of our loglik models
                loglik_mean, loglik_mode = read_functions.statistics(p_params_complete['x'],self.types_list)
                #            loglik_mean[np.isinf(loglik_mean)] = 1e20

                # print('train_data_transformed',type(train_data_transformed))
                # print('train_data_transformed',train_data_transformed.shape)
                # print('loglik_mean',type(loglik_mean))
                # print('loglik_mean',loglik_mean.shape)
                # print('miss_mask_aux',type(miss_mask_aux))
                # print('miss_mask_aux',miss_mask_aux.shape)
                # print('n_batches',n_batches)
                        
                #Try this for the errors
                #ak: lost examples when batch_size and number of variables are not modulus 0
                error_train_mean, error_test_mean = read_functions.error_computation(train_data_transformed, loglik_mean, self.types_list, miss_mask_aux[:n_batches*batch_size_here,:])
                error_train_mode, error_test_mode = read_functions.error_computation(train_data_transformed, loglik_mode, self.types_list, miss_mask_aux[:n_batches*batch_size_here,:])
                error_train_samples, error_test_samples = read_functions.error_computation(train_data_transformed, est_data_transformed, self.types_list, miss_mask_aux[:n_batches*batch_size_here,:])
                error_train_imputed, error_test_imputed = read_functions.error_computation(train_data_transformed, est_data_imputed, self.types_list, miss_mask_aux[:n_batches*batch_size_here,:])

                #Compute test-loglik from log_p_x_missing
                log_p_x_total = np.transpose(np.concatenate(log_p_x_total,1))
                log_p_x_missing_total = np.transpose(np.concatenate(log_p_x_missing_total,1))
                if true_miss_set:
                    log_p_x_missing_total = np.multiply(log_p_x_missing_total,true_miss_mask_aux[:n_batches*batch_size_here,:])
                avg_test_loglik = 100000  ##  np.finfo(float).max #ak - maximal float number = ugly for printing
                if miss_set:
                    avg_test_loglik = np.sum(log_p_x_missing_total)/np.sum(1.0-miss_mask_aux)

                # Display logs per epoch step
                if self.display:
                    self.print_loss(epoch, start_time, avg_loss/n_batches, avg_test_loglik, avg_KL_s/n_batches, avg_KL_z/n_batches)
                    # print('Test error mode: ' + str(np.round(np.mean(error_test_mode),3)))
                    # print("")
                
                #Compute train and test loglik per variables
                loglik_per_variable = np.sum(log_p_x_total,0)/np.sum(miss_mask_aux,0)

                #ak: only compute if a missing file was supplied
                loglik_per_variable_missing =  100000  ##  np.finfo(float).max #ak - maximal float number = ugly for printing
                if miss_set:
                    loglik_per_variable_missing = np.sum(log_p_x_missing_total,0)/np.sum(1.0-miss_mask_aux,0)
            
                #Store evolution of all the terms in the ELBO
                loglik_epoch.append(loglik_per_variable)
                testloglik_epoch.append(loglik_per_variable_missing)
                KL_s_epoch.append(avg_KL_s/n_batches)
                KL_z_epoch.append(avg_KL_z/n_batches)
                error_train_mode_global.append(error_train_mode)
                error_test_mode_global.append(error_test_mode)
            
            
                # if epoch % self.save == 0:
                #     self.vprint(1,'Saving Variables ...')  
                #     save_path = saver.save(session, self.network_file_name)
                
            if training_phase:
                self.vprint(1,'Training Finished ...')

                # could be saved only if required
                self.vprint(2,'Saving informations ...')
                for (data_file_partial_name,data_values) in [
                        ('loglik',loglik_epoch),
                        ('KL_s',np.reshape(KL_s_epoch,[-1,1])),
                        ('KL_z',np.reshape(KL_z_epoch,[-1,1])),
                        ('train_error',error_train_mode_global),
                        ('test_error',error_test_mode_global),
                        #ak: hack for now - as entries are not lists 
                        ('testloglik',[[x] for x in testloglik_epoch]),
                        ]:
                    self.save_data(self.full_results_path,'{}_{}.csv'.format(self.experiment_name,data_file_partial_name),data_values)

                # self.save_data(self.full_results_path,'{}_loglik.csv'.format(self.experiment_name),loglik_epoch)
                # self.save_data(self.full_results_path,'{}_KL_s.csv'.format(self.experiment_name),np.reshape(KL_s_epoch,[-1,1]))
                # self.save_data(self.full_results_path,'{}_KL_z.csv'.format(self.experiment_name),np.reshape(KL_z_epoch,[-1,1]))
                # self.save_data(self.full_results_path,'{}_train_error.csv'.format(self.experiment_name),error_train_mode_global)
                # self.save_data(self.full_results_path,'{}_test_error.csv'.format(self.experiment_name),error_test_mode_global)
                # #ak: hack for now - as entries are not lists 
                # self.save_data(self.full_results_path,'{}_testloglik.csv'.format(self.experiment_name),[[x] for x in testloglik_epoch])

                # Save the network/variables to disk at the end
                self.vprint(1,'Saving Network ... <<{}>>'.format(self.network_file_name))  
                save_path = saver.save(session, self.network_file_name)
                # not sure if this is required
                #self.save_path = save_path
                # print('save_path',save_path)
            
            elif testing_phase:
                self.vprint(1,'Testing Finished ...')
                #Compute the data reconstruction
            
                data_reconstruction = train_data_transformed * miss_mask_aux[:n_batches*batch_size_here,:] + \
                  np.round(loglik_mode,3) * (1 - miss_mask_aux[:n_batches*batch_size_here,:])
                  
                train_data_transformed = train_data_transformed[np.argsort(random_perm)]
                loglik_mean_reconstructed = loglik_mean[np.argsort(random_perm)]
                data_reconstruction = data_reconstruction[np.argsort(random_perm)]

                self.vprint(2,'Saving reconstructions ...')
                for (data_file_partial_name,data_values) in [
                        ('data_reconstruction',data_reconstruction),
                        ('data_true',train_data_transformed),
                        ('data_loglik_mean_reconstructed',loglik_mean_reconstructed),
                        ]:
                    self.save_data(self.full_results_path,'{}_{}.csv'.format(self.experiment_name,data_file_partial_name),data_values)

                # self.save_data(self.full_results_path,'{}_data_reconstruction.csv'.format(self.experiment_name),data_reconstruction)
                # self.save_data(self.full_results_path,'{}_data_true.csv'.format(self.experiment_name),train_data_transformed)
                # self.save_data(self.full_results_path,'{}_data_loglik_mean_reconstructed.csv'.format(self.experiment_name),loglik_mean_reconstructed)

                df_real        = pd.DataFrame(train_data_transformed)
                df_loglik_mean = pd.DataFrame(loglik_mean_reconstructed)
                self.vprint(2,'Reconstruction Correlation:')
                self.vprint(2,df_real.corrwith(df_loglik_mean))
                

                return (train_data_transformed,data_reconstruction,loglik_mean_reconstructed,
                            z_total[np.argsort(random_perm)],
                            s_total[np.argsort(random_perm)])

        
            






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Dec  6 15:40:14 2017
Functions needed to read the data from different databases
@author: anazabal, olmosUC3M, ivaleraM
"""

import csv
import numpy as np
import os
import scipy.io as sc
from sklearn.metrics import mean_squared_error
from pathlib import Path

def saveInt(potentialInt):
    intReturn = None
    try:
        intReturn = int(potentialInt)
    except:
        intReturn = 0
        pass
    return intReturn

def get_types_list(types_description):
    types_list = None
    # test whether the types are passed as a list - if not assume it is a file which was supplied
    if type(types_description) != type([]):
        #Read the types of the data from the files
        with open(types_description) as f:
            types_list = [{k: v for k, v in row.items()}
            for row in csv.DictReader(f, skipinitialspace=True)]
    else:
        try:
            types_description[0]['type']
            types_list = types_description
        except:
            types_list = [{'attribute_name':str(x[0]),'type':str(x[1]),'dim':str(x[2]),'nclass': x[3] and str(x[3]) or ''} for x in types_description]
            pass
    return types_list


def read_data_df(data_df,types_description,true_missing_mask=None):
    # moved to extra method
    types_list = get_types_list(types_description)
    data = np.array(data_df)


    # handle truely missing data
    if type(true_missing_mask) != type(None):
        data_masked = np.ma.masked_where(np.isnan(data),data)
        #We need to fill the data depending on the given data...
        data_filler = []
        for i in range(len(types_list)):
            if types_list[i]['type'] == 'cat' or types_list[i]['type'] == 'ordinal':
                aux = np.unique(data[:,i])
                data_filler.append(aux[0])  #Fill with the first element of the cat (0, 1, or whatever)
            else:
                data_filler.append(0.0)

        data = data_masked.filled(data_filler)
        
    # Construct the data matrices
    data_complete = []
    for i in range(np.shape(data)[1]):
        if types_list[i]['type'] == 'cat':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_list[i]['dim']))
            cat_data = new_categories[indexes]
            # Create one hot encoding for the categories
            aux = np.zeros([np.shape(data)[0], len(new_categories)])
            aux[np.arange(np.shape(data)[0]), cat_data] = 1
            data_complete.append(aux)

        elif types_list[i]['type'] == 'ordinal':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_list[i]['dim']))
            cat_data = new_categories[indexes]
            # Create thermometer encoding for the categories
            aux = np.zeros([np.shape(data)[0], 1 + len(new_categories)])
            aux[:, 0] = 1
            aux[np.arange(np.shape(data)[0]), 1 + cat_data] = -1
            aux = np.cumsum(aux, 1)
            data_complete.append(aux[:, :-1])

        elif types_list[i]['type'] == 'count':
            if np.min(data[:, i]) == 0:
                aux = data[:, i] + 1
                data_complete.append(np.transpose([aux]))
            else:
                data_complete.append(np.transpose([data[:, i]]))
        else:
            data_complete.append(np.transpose([data[:, i]]))

    data = np.concatenate(data_complete, 1)
    n_samples = np.shape(data)[0]

    return data, types_list, n_samples
    

def read_data_df_as_input(datadf, types_description, miss_file, true_miss):

    # moved to extra method
    types_list = get_types_list(types_description)
    
    data = np.array(datadf)
    # Sustitute NaN values by something (we assume we have the real missing value mask)
    if true_miss:
        missing_positions = np.array(true_miss)
        true_miss_mask = np.ones([np.shape(data)[0], len(types_list)])
        true_miss_mask[missing_positions[:, 0] - 1, missing_positions[:, 1] - 1] = 0  # Indexes in the csv start at 1
        data_masked = np.ma.masked_where(np.isnan(data), data)
        # We need to fill the data depending on the given data...
        data_filler = []
        for i in range(len(types_list)):
            if types_list[i]['type'] == 'cat' or types_list[i]['type'] == 'ordinal':
                aux = np.unique(data[:, i])
                data_filler.append(aux[0])  # Fill with the first element of the cat (0, 1, or whatever)
            else:
                data_filler.append(0.0)

        data = data_masked.filled(data_filler)
    else:
        true_miss_mask = np.ones([np.shape(data)[0], len(types_list)])  # It doesn't affect our data

    # Construct the data matrices
    data_complete = []
    for i in range(np.shape(data)[1]):
        if types_list[i]['type'] == 'cat':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_list[i]['dim']))
            cat_data = new_categories[indexes]
            # Create one hot encoding for the categories
            aux = np.zeros([np.shape(data)[0], len(new_categories)])
            aux[np.arange(np.shape(data)[0]), cat_data] = 1
            data_complete.append(aux)

        elif types_list[i]['type'] == 'ordinal':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_list[i]['dim']))
            cat_data = new_categories[indexes]
            # Create thermometer encoding for the categories
            aux = np.zeros([np.shape(data)[0], 1 + len(new_categories)])
            aux[:, 0] = 1
            aux[np.arange(np.shape(data)[0]), 1 + cat_data] = -1
            aux = np.cumsum(aux, 1)
            data_complete.append(aux[:, :-1])

        elif types_list[i]['type'] == 'count':
            if np.min(data[:, i]) == 0:
                aux = data[:, i] + 1
                data_complete.append(np.transpose([aux]))
            else:
                data_complete.append(np.transpose([data[:, i]]))



        else:
            data_complete.append(np.transpose([data[:, i]]))

    data = np.concatenate(data_complete, 1)

    # Read Missing mask from csv (contains positions of missing values)
    n_samples = np.shape(data)[0]
    n_variables = len(types_list)
    miss_mask = np.ones([np.shape(data)[0], n_variables])
    # If there is no mask, assume all data is observed
    if miss_file and os.path.isfile(miss_file):
        with open(miss_file, 'r') as f:
            missing_positions = [[int(x) for x in rec] for rec in csv.reader(f, delimiter=',')]
            missing_positions = np.array(missing_positions)
        miss_mask[missing_positions[:, 0] - 1, missing_positions[:, 1] - 1] = 0  # Indexes in the csv start at 1

    return data, types_list, miss_mask, true_miss_mask, n_samples

def read_data(data_file, types_file, miss_file, true_miss_file):
    # Read types of data from data file
    with open(types_file) as f:
        types_dict = [{k: v for k, v in row.items()}
                      for row in csv.DictReader(f, skipinitialspace=True)]

    # Read data from input file
    with open(data_file, 'r') as f:
        data = [[float(x) for x in rec] for rec in csv.reader(f, delimiter=',')]
        data = np.array(data)

    # Sustitute NaN values by something (we assume we have the real missing value mask)
    if true_miss_file:
        with open(true_miss_file, 'r') as f:
            missing_positions = [[int(x) for x in rec] for rec in csv.reader(f, delimiter=',')]
            missing_positions = np.array(missing_positions)

        true_miss_mask = np.ones([np.shape(data)[0], len(types_dict)])
        true_miss_mask[missing_positions[:, 0] - 1, missing_positions[:, 1] - 1] = 0  # Indexes in the csv start at 1
        data_masked = np.ma.masked_where(np.isnan(data), data)
        # We need to fill the data depending on the given data...
        data_filler = []
        for i in range(len(types_dict)):
            if types_dict[i]['type'] == 'cat' or types_dict[i]['type'] == 'ordinal':
                aux = np.unique(data[:, i])
                data_filler.append(aux[0])  # Fill with the first element of the cat (0, 1, or whatever)
            else:
                data_filler.append(0.0)

        data = data_masked.filled(data_filler)
    else:
        true_miss_mask = np.ones([np.shape(data)[0], len(types_dict)])  # It doesn't affect our data

    # Construct the data matrices
    data_complete = []
    for i in range(np.shape(data)[1]):

        if types_dict[i]['type'] == 'cat':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_dict[i]['dim']))
            cat_data = new_categories[indexes]
            # Create one hot encoding for the categories
            aux = np.zeros([np.shape(data)[0], len(new_categories)])
            aux[np.arange(np.shape(data)[0]), cat_data] = 1
            data_complete.append(aux)

        elif types_dict[i]['type'] == 'ordinal':
            # Get categories
            cat_data = [int(x) for x in data[:, i]]
            categories, indexes = np.unique(cat_data, return_inverse=True)
            # Transform categories to a vector of 0:n_categories
            new_categories = np.arange(int(types_dict[i]['dim']))
            cat_data = new_categories[indexes]
            # Create thermometer encoding for the categories
            aux = np.zeros([np.shape(data)[0], 1 + len(new_categories)])
            aux[:, 0] = 1
            aux[np.arange(np.shape(data)[0]), 1 + cat_data] = -1
            aux = np.cumsum(aux, 1)
            data_complete.append(aux[:, :-1])

        elif types_dict[i]['type'] == 'count':
            if np.min(data[:, i]) == 0:
                aux = data[:, i] + 1
                data_complete.append(np.transpose([aux]))
            else:
                data_complete.append(np.transpose([data[:, i]]))



        else:
            data_complete.append(np.transpose([data[:, i]]))

    data = np.concatenate(data_complete, 1)

    # Read Missing mask from csv (contains positions of missing values)
    n_samples = np.shape(data)[0]
    n_variables = len(types_dict)
    miss_mask = np.ones([np.shape(data)[0], n_variables])
    # If there is no mask, assume all data is observed
    if os.path.isfile(miss_file):
        with open(miss_file, 'r') as f:
            missing_positions = [[int(x) for x in rec] for rec in csv.reader(f, delimiter=',')]
            missing_positions = np.array(missing_positions)
        miss_mask[missing_positions[:, 0] - 1, missing_positions[:, 1] - 1] = 0  # Indexes in the csv start at 1

    return data, types_dict, miss_mask, true_miss_mask, n_samples


def next_batch(data, types_dict, miss_mask, batch_size, index_batch):
    # Create minibath
    batch_xs = data[index_batch * batch_size:(index_batch + 1) * batch_size, :]

    # Slipt variables of the batches
    data_list = []
    initial_index = 0
    for d in types_dict:
        dim = int(d['dim'])
        data_list.append(batch_xs[:, initial_index:initial_index + dim])
        initial_index += dim

    # Missing data
    miss_list = miss_mask[index_batch * batch_size:(index_batch + 1) * batch_size, :]

    return data_list, miss_list


def samples_concatenation(samples):
    for i, batch in enumerate(samples):
        #print('DEBUG', i,batch)
        if i == 0:
            samples_x = np.concatenate(batch['x'], 1)
            samples_y = batch['y']
            samples_z = batch['z']
            samples_s = batch['s']
        else:
            samples_x = np.concatenate([samples_x, np.concatenate(batch['x'], 1)], 0)
            samples_y = np.concatenate([samples_y, batch['y']], 0)
            samples_z = np.concatenate([samples_z, batch['z']], 0)
            samples_s = np.concatenate([samples_s, batch['s']], 0)

    return samples_s, samples_z, samples_y, samples_x


def discrete_variables_transformation(data, types_dict):
    ind_ini = 0
    output = []
    for d in range(len(types_dict)):
        ind_end = ind_ini + int(types_dict[d]['dim'])
        if types_dict[d]['type'] == 'cat':
            output.append(np.reshape(np.argmax(data[:, ind_ini:ind_end], 1), [-1, 1]))
        elif types_dict[d]['type'] == 'ordinal':
            output.append(np.reshape(np.sum(data[:, ind_ini:ind_end], 1) - 1, [-1, 1]))
        else:
            output.append(data[:, ind_ini:ind_end])
        ind_ini = ind_end

    return np.concatenate(output, 1)


# Several baselines
def mean_imputation(train_data, miss_mask, types_dict):
    ind_ini = 0
    est_data = []
    for dd in range(len(types_dict)):
        # Imputation for cat and ordinal is done using the mode of the data
        if types_dict[dd]['type'] == 'cat' or types_dict[dd]['type'] == 'ordinal':
            ind_end = ind_ini + 1
            # The imputation is based on whatever is observed
            miss_pattern = (miss_mask[:, dd] == 1)
            values, counts = np.unique(train_data[miss_pattern, ind_ini:ind_end], return_counts=True)
            data_mode = np.argmax(counts)
            data_imputed = train_data[:, ind_ini:ind_end] * miss_mask[:, ind_ini:ind_end] + data_mode * (
                        1.0 - miss_mask[:, ind_ini:ind_end])

        # Imputation for the rest of the variables is done with the mean of the data
        else:
            ind_end = ind_ini + int(types_dict[dd]['dim'])
            miss_pattern = (miss_mask[:, dd] == 1)
            # The imputation is based on whatever is observed
            data_mean = np.mean(train_data[miss_pattern, ind_ini:ind_end], 0)
            data_imputed = train_data[:, ind_ini:ind_end] * miss_mask[:, ind_ini:ind_end] + data_mean * (
                        1.0 - miss_mask[:, ind_ini:ind_end])

        est_data.append(data_imputed)
        ind_ini = ind_end

    return np.concatenate(est_data, 1)


def p_distribution_params_concatenation(params, types_dict, z_dim, s_dim):
    keys = params[0].keys()
    out_dict = {key: [] for key in keys}

    for i, batch in enumerate(params):

        for d, k in enumerate(keys):

            if k == 'z' or k == 'y':
                if i == 0:
                    out_dict[k] = batch[k]
                else:
                    out_dict[k] = np.concatenate([out_dict[k], batch[k]], 1)

            elif k == 'x':
                if i == 0:
                    out_dict[k] = batch[k]
                else:
                    for v in range(len(types_dict)):
                        if types_dict[v]['type'] == 'pos' or types_dict[v]['type'] == 'real':
                            out_dict[k][v] = np.concatenate([out_dict[k][v], batch[k][v]], 1)
                        else:
                            out_dict[k][v] = np.concatenate([out_dict[k][v], batch[k][v]], 0)

    return out_dict


def q_distribution_params_concatenation(params, z_dim, s_dim):
    keys = params[0].keys()
    out_dict = {key: [] for key in keys}

    for i, batch in enumerate(params):
        for d, k in enumerate(keys):
            out_dict[k].append(batch[k])

    out_dict['z'] = np.concatenate(out_dict['z'], 1)
    if 's' in out_dict:
        out_dict['s'] = np.concatenate(out_dict['s'], 0)

    return out_dict


def statistics(loglik_params, types_dict):
    # print('AK-DEBUG-{}'.format(Path(__file__).stem))
    # print('loglik_params',len(loglik_params[0]))
    # print('AK-DEBUG-{}\n'.format(Path(__file__).stem))
    loglik_mean = []
    loglik_mode = []

    for d, attrib in enumerate(loglik_params):
        # print('AK-DEBUG-{}'.format(Path(__file__).stem))
        # print('loglik_mean',len(loglik_mean))
        # print('AK-DEBUG-{}\n'.format(Path(__file__).stem))
        if types_dict[d]['type'] == 'real':
            # Normal distribution (mean, sigma)
            loglik_mean.append(attrib[0])
            loglik_mode.append(attrib[0])
        # Only for log-normal
        elif types_dict[d]['type'] == 'pos':
            # Log-normal distribution (mean, sigma)
            loglik_mean.append(np.maximum(np.exp(attrib[0] + 0.5 * attrib[1]) - 1.0, 0.0))
            loglik_mode.append(np.maximum(np.exp(attrib[0] - attrib[1]) - 1.0, 0.0))
        elif types_dict[d]['type'] == 'count':
            # Poisson distribution (lambda)
            loglik_mean.append(attrib)
            loglik_mode.append(np.floor(attrib))

        else:
            # Categorical and ordinal (mode imputation for both)
            loglik_mean.append(np.reshape(np.argmax(attrib, 1), [-1, 1]))
            loglik_mode.append(np.reshape(np.argmax(attrib, 1), [-1, 1]))

    # print('AK-DEBUG-{}'.format(Path(__file__).stem))
    # print('loglik_mean',len(loglik_mean))
    # print('squeeze loglik_mean',len(np.squeeze(loglik_mean)))
    # print('AK-DEBUG-{}\n'.format(Path(__file__).stem))

    return np.transpose(np.squeeze(loglik_mean)), np.transpose(np.squeeze(loglik_mode))


def error_computation(x_train, x_hat, types_dict, miss_mask):
    error_observed = []
    error_missing = []
    ind_ini = 0
    for dd in range(len(types_dict)):
        # Mean classification error
        if types_dict[dd]['type'] == 'cat':
            ind_end = ind_ini + 1
            error_observed.append(np.mean(
                x_train[miss_mask[:, dd] == 1, ind_ini:ind_end] != x_hat[miss_mask[:, dd] == 1, ind_ini:ind_end]))
            if np.sum(miss_mask[:, dd] == 0, 0) == 0:
                error_missing.append(0)
            else:
                error_missing.append(np.mean(
                    x_train[miss_mask[:, dd] == 0, ind_ini:ind_end] != x_hat[miss_mask[:, dd] == 0, ind_ini:ind_end]))
        # Mean "shift" error
        elif types_dict[dd]['type'] == 'ordinal':
            ind_end = ind_ini + 1
            error_observed.append(np.mean(np.abs(
                x_train[miss_mask[:, dd] == 1, ind_ini:ind_end] - x_hat[miss_mask[:, dd] == 1, ind_ini:ind_end])) / int(
                types_dict[dd]['dim']))
            if np.sum(miss_mask[:, dd] == 0, 0) == 0:
                error_missing.append(0)
            else:
                error_missing.append(np.mean(np.abs(
                    x_train[miss_mask[:, dd] == 0, ind_ini:ind_end] - x_hat[miss_mask[:, dd] == 0,
                                                                      ind_ini:ind_end])) / int(types_dict[dd]['dim']))
        # Normalized root mean square error
        else:
            ind_end = ind_ini + int(types_dict[dd]['dim'])
            norm_term = np.max(x_train[:, dd]) - np.min(x_train[:, dd])
            error_observed.append(np.sqrt(mean_squared_error(x_train[miss_mask[:, dd] == 1, ind_ini:ind_end],
                                                             x_hat[miss_mask[:, dd] == 1,
                                                             ind_ini:ind_end])) / norm_term)

            if np.sum(miss_mask[:, dd] == 0, 0) == 0:
                error_missing.append(0)
            else:
                error_missing.append(np.sqrt(mean_squared_error(x_train[miss_mask[:, dd] == 0, ind_ini:ind_end],
                                                                x_hat[miss_mask[:, dd] == 0,
                                                                ind_ini:ind_end])) / norm_term)

        ind_ini = ind_end

    return error_observed, error_missing



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 23 15:49:15 2018

Graph definition for all models

@author: anazabal, olmosUC3M, ivaleraM
@changes: athro
"""

import tensorflow as tf
import numpy as np
from hivae import VAE_functions

def HVAE_graph(model_name, types_description, batch_size, learning_rate=1e-3, z_dim=2, y_dim=1, s_dim=2, y_dim_partition=[]):
    
    #We select the model for the VAE
    # print('\n'*5)
    print('[*] Importing model: ' + model_name)
    # how to import a submoule 101
    model = __import__('hivae.{}'.format(model_name),fromlist=[model_name])
    #model = __import__(model_name)
    
    #Load placeholders
    print('[*] Defining placeholders')
    batch_data_list, batch_data_list_observed, miss_list, tau, tau2, types_list = VAE_functions.place_holder_types(types_description, batch_size)
    
    #Batch normalization of the data
    X_list, normalization_params = VAE_functions.batch_normalization(batch_data_list_observed, types_list, miss_list)
    
    #Set dimensionality of Y
    if y_dim_partition:
        y_dim_output = np.sum(y_dim_partition)
    else:
        y_dim_partition = y_dim*np.ones(len(types_list),dtype=int)
        y_dim_output = np.sum(y_dim_partition)
    
    #Encoder definition
    print('[*] Defining Encoder...')
    samples, q_params = model.encoder(X_list, miss_list, batch_size, z_dim, s_dim, tau)
    encoding = samples
    # print('debug ' * 10)
    # print(samples)
    # print('- ' * 10)
    # print(q_params)
    # print('debug ' * 10)

    
    print('[*] Defining Decoder...')
    theta, samples, p_params, log_p_x, log_p_x_missing = model.decoder(batch_data_list, miss_list, types_list, samples, q_params, normalization_params, batch_size, z_dim, y_dim_output, y_dim_partition, tau2)

    print('[*] Defining Cost function...')
    ELBO, loss_reconstruction, KL_z, KL_s = model.cost_function(log_p_x, p_params, q_params, types_list, z_dim, y_dim_output, s_dim)
    
    optim = tf.train.AdamOptimizer(learning_rate).minimize(-ELBO)
    
    #Generator function for testing purposes
    samples_test, test_params, log_p_x_test, log_p_x_missing_test = model.samples_generator(batch_data_list, X_list, miss_list, types_list, batch_size, z_dim, y_dim_output, y_dim_partition, s_dim, tau, tau2, normalization_params)
    
    #Packing results
    tf_nodes = {'ground_batch' : batch_data_list,
                'ground_batch_observed' : batch_data_list_observed,
                'miss_list': miss_list,
                'tau_GS': tau,
                'tau_var': tau2,
                'samples': samples,
                'log_p_x': log_p_x,
                'log_p_x_missing': log_p_x_missing,
                'loss_re' : loss_reconstruction,
                'loss': -ELBO, 
                'optim': optim,
                'KL_s': KL_s,
                'KL_z': KL_z,
                'p_params': p_params,
                'q_params': q_params,
                'samples_test': samples_test,
                'test_params': test_params,
                'log_p_x_test': log_p_x,
                'log_p_x_missing_test': log_p_x_missing_test,
                'encoding':encoding}

    return tf_nodes



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 10:59:14 2018

@author: anazabal
"""

import csv
import tensorflow as tf
from hivae import loglik_models_missing_normalize
import numpy as np


def place_holder_types(types_description, batch_size):

    #print(types_description)
    types_list = None
    # test whether the types are passed as a dlist - if not assume it is a file which was supplied
    if type(types_description) != type([]):
        
        #Read the types of the data from the files
        with open(types_description) as f:
            types_list = [{k: v for k, v in row.items()}
            for row in csv.DictReader(f, skipinitialspace=True)]
    # else:
    #     types_list = [{'type':str(x[1]),'dim':str(x[2]),'nclass': x[3] and str(x[3]) or ''} for x in types_description]
    # for now - assume that we already have a list with the parsed information
    else:
        types_list = types_description
        
            
    #Create placeholders for every data type, with appropriate dimensions
    batch_data_list = []
    for i in range(len(types_list)):
        batch_data_list.append(tf.placeholder(tf.float32, shape=(batch_size,types_list[i]['dim'])))
    tf.concat(batch_data_list, axis=1)
    
    #Create placeholders for every missing data type, with appropriate dimensions
    batch_data_list_observed = []
    for i in range(len(types_list)):
        batch_data_list_observed.append(tf.placeholder(tf.float32, shape=(batch_size,types_list[i]['dim'])))
    tf.concat(batch_data_list_observed, axis=1)
        
    #Create placeholders for the missing data indicator variable
    miss_list = tf.placeholder(tf.int32, shape=(batch_size,len(types_list)))
    
    #Placeholder for Gumbel-softmax parameter
    tau = tf.placeholder(tf.float32,shape=())
    tau2 = tf.placeholder(tf.float32,shape=())
    # print('debug ' * 10)
    # print(batch_data_list, batch_data_list_observed, miss_list, tau, tau2, types_list)
    # print('debug ' * 10)
    return batch_data_list, batch_data_list_observed, miss_list, tau, tau2, types_list

def batch_normalization(batch_data_list, types_list, miss_list):
    
    normalized_data = []
    normalization_parameters = []
    
    for i,d in enumerate(batch_data_list):
        #Partition the data in missing data (0) and observed data n(1)
        missing_data, observed_data = tf.dynamic_partition(d, miss_list[:,i], num_partitions=2)
        condition_indices = tf.dynamic_partition(tf.range(tf.shape(d)[0]), miss_list[:,i], num_partitions=2)
        
        if types_list[i]['type'] == 'real':
            #We transform the data to a gaussian with mean 0 and std 1
            data_mean, data_var = tf.nn.moments(observed_data,0)
            data_var = tf.clip_by_value(data_var,1e-6,1e20) #Avoid zero values
            aux_X = tf.nn.batch_normalization(observed_data,data_mean,data_var,offset=0.0,scale=1.0,variance_epsilon=1e-6)
            
            normalized_data.append(tf.dynamic_stitch(condition_indices, [missing_data, aux_X]))
            normalization_parameters.append([data_mean, data_var])
            
        #When using log-normal
        elif types_list[i]['type'] == 'pos':
#           #We transform the log of the data to a gaussian with mean 0 and std 1
            observed_data_log = tf.log(1.0 + observed_data)
            data_mean_log, data_var_log = tf.nn.moments(observed_data_log,0)
            
            data_var_log = tf.clip_by_value(data_var_log,1e-6,1e20) #Avoid zero values
            aux_X = tf.nn.batch_normalization(observed_data_log,data_mean_log,data_var_log,offset=0.0,scale=1.0,variance_epsilon=1e-6)
            
            normalized_data.append(tf.dynamic_stitch(condition_indices, [missing_data, aux_X]))
            normalization_parameters.append([data_mean_log, data_var_log])
            
        elif types_list[i]['type'] == 'count':
            
            #Input log of the data
            aux_X = tf.log(observed_data)
            
            normalized_data.append(tf.dynamic_stitch(condition_indices, [missing_data, aux_X]))
            normalization_parameters.append([0.0, 1.0])
            
        else:
            #Don't normalize the categorical and ordinal variables
            normalized_data.append(d)
            normalization_parameters.append([0.0, 1.0]) #No normalization here
    
    return normalized_data, normalization_parameters


def s_proposal_multinomial(X, batch_size, s_dim, tau, reuse):
    
    #We propose a categorical distribution to create a GMM for the latent space z
    log_pi = tf.layers.dense(inputs=X, units=s_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'enc_s', reuse=reuse)
    
    
    #Gumbel-softmax trick
    log_pi_aux = tf.log(tf.clip_by_value(tf.nn.softmax(log_pi),1e-6,1))
    U = -tf.log(-tf.log(tf.random_uniform([batch_size,s_dim])))
    samples_s = tf.nn.softmax((log_pi_aux + U)/tau)
    
    return samples_s, log_pi_aux

def z_proposal_GMM(X, samples_s, batch_size, z_dim, reuse):
    
#    X_in = tf.layers.dense(inputs=X, units=100, activation=tf.nn.tanh,
#                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_0_' + 'mean_enc_z', reuse=reuse)
    
    #We propose a GMM for z
    mean_qz = tf.layers.dense(inputs=tf.concat([X,samples_s],1), units=z_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'mean_enc_z', reuse=reuse)
    log_var_qz = tf.layers.dense(inputs=tf.concat([X,samples_s],1), units=z_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'logvar_enc_z', reuse=reuse)
    
    # Avoid numerical problems
    log_var_qz = tf.clip_by_value(log_var_qz,-15.0,15.0)
    # Rep-trick
    eps = tf.random_normal((batch_size, z_dim), 0, 1, dtype=tf.float32)
    samples_z = mean_qz+tf.multiply(tf.exp(log_var_qz/2), eps)
    
    return samples_z, [mean_qz, log_var_qz]

def z_proposal_Normal(X, batch_size, z_dim, reuse):
    
    #We propose a GMM for z
    mean_qz = tf.layers.dense(inputs=X, units=z_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'mean_enc_z', reuse=reuse)
    log_var_qz = tf.layers.dense(inputs=X, units=z_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'logvar_enc_z', reuse=reuse)
    
    # Avoid numerical problems
    log_var_qz = tf.clip_by_value(log_var_qz,-15.0,15.0)
    # Rep-trick
    eps = tf.random_normal((batch_size, z_dim), 0, 1, dtype=tf.float32)
    samples_z = mean_qz+tf.multiply(tf.exp(log_var_qz/2), eps)
    
    return samples_z, [mean_qz, log_var_qz]

def z_proposal_GMM_factorized(X, samples_s, miss_list, batch_size, z_dim, reuse):
    
    mean_qz = []
    log_var_qz = []
    
    for i,d in enumerate(X):
        
        #Partition the data in missing data (0) and observed data n(1)
        missing_data, observed_data = tf.dynamic_partition(d, miss_list[:,i], num_partitions=2)
        missing_s, observed_s = tf.dynamic_partition(samples_s, miss_list[:,i], num_partitions=2)
        condition_indices = tf.dynamic_partition(tf.range(tf.shape(d)[0]), miss_list[:,i], num_partitions=2)
        
        #Get the dimensions of the observed data
        nObs = tf.shape(observed_data)[0]
        
        #Mean layer
        aux_m = tf.layers.dense(inputs=tf.concat([observed_data,observed_s],1), units=z_dim, activation=None,
                             kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'mean_enc_z'+str(i), reuse=reuse)
        #Reconstruct means with zeros (so they don't affect the mean_joint)
        aux_mean_qz = tf.dynamic_stitch(condition_indices, [tf.zeros([batch_size-nObs,z_dim],dtype=tf.float32),aux_m])

        #Logvar layers
        aux_lv = tf.layers.dense(inputs=tf.concat([observed_data,observed_s],1), units=z_dim, activation=None,
                             kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_' + 'logvar_enc_z'+str(i), reuse=reuse)
        #Set a high value to make the variance in the missing cases negligible
        aux_log_var_qz = tf.dynamic_stitch(condition_indices, [tf.fill([batch_size-nObs,z_dim],15.0),aux_lv])
        
        mean_qz.append(aux_mean_qz)
        log_var_qz.append(aux_log_var_qz)
        
    #Input prior
    log_var_qz.append(tf.zeros([batch_size,z_dim]))
    mean_qz.append(tf.zeros([batch_size,z_dim]))
    #Compute full parameters, as a product of Gaussians distribution
    log_var_qz_joint = -tf.reduce_logsumexp(tf.negative(log_var_qz), 0)
    mean_qz_joint = tf.multiply(tf.exp(log_var_qz_joint), tf.reduce_sum(tf.multiply(mean_qz,tf.exp(tf.negative(log_var_qz))), 0))
    
    # Avoid numerical problems
    log_var_qz = tf.clip_by_value(log_var_qz,-15.0,15.0)
    # Rep-trick
    eps = tf.random_normal((batch_size, z_dim), 0, 1, dtype=tf.float32)
    samples_z = mean_qz_joint+tf.multiply(tf.exp(log_var_qz_joint/2), eps)
    
    return samples_z, [mean_qz_joint, log_var_qz_joint]

def z_distribution_GMM(samples_s, z_dim, reuse):
    
    #We propose a GMM for z
    mean_pz = tf.layers.dense(inputs=samples_s, units=z_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05), name= 'layer_1_' + 'mean_dec_z', reuse=reuse)
    log_var_pz = tf.zeros([tf.shape(samples_s)[0],z_dim])
    
    # Avoid numerical problems
    log_var_pz = tf.clip_by_value(log_var_pz,-15.0,15.0)
    
    return mean_pz, log_var_pz

def y_partition(samples_y, types_list, y_dim_partition):
    # print(y_dim_partition)
    # print(types_list)
    grouped_samples_y = []
    #First element must be 0 and the length of the partition vector must be len(types_dict)+1
    if len(y_dim_partition) != len(types_list):
        raise Exception("The length of the partition vector must match the number of variables in the data + 1")
        
    #Insert a 0 at the beginning of the cumsum vector
    partition_vector_cumsum = np.insert(np.cumsum(y_dim_partition),0,0)
    for i in range(len(types_list)):
        grouped_samples_y.append(samples_y[:,partition_vector_cumsum[i]:partition_vector_cumsum[i+1]])
    
    return grouped_samples_y

def theta_estimation_from_z(samples_z, types_list, miss_list, batch_size, reuse):
    
    theta = []
    
    #Independet yd -> Compute p(xd|yd)
    for i,d in enumerate(types_list):
        
        #Partition the data in missing data (0) and observed data (1)
        missing_y, observed_y = tf.dynamic_partition(samples_z, miss_list[:,i], num_partitions=2)
        condition_indices = tf.dynamic_partition(tf.range(tf.shape(samples_z)[0]), miss_list[:,i], num_partitions=2)
        nObs = tf.shape(observed_y)[0]
        
        #Different layer models for each type of variable
        if types_list[i]['type'] == 'real':
            params = theta_real(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'pos':
            params = theta_pos(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'count':
            params = theta_count(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'cat':
            params = theta_cat(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'ordinal':
            params = theta_ordinal(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        theta.append(params)
            
    return theta

def theta_estimation_from_y(samples_y, types_list, miss_list, batch_size, reuse):
    
    theta = []
    
    #Independet yd -> Compute p(xd|yd)
    for i,d in enumerate(samples_y):
        
        #Partition the data in missing data (0) and observed data (1)
        missing_y, observed_y = tf.dynamic_partition(d, miss_list[:,i], num_partitions=2)
        condition_indices = tf.dynamic_partition(tf.range(tf.shape(d)[0]), miss_list[:,i], num_partitions=2)
        nObs = tf.shape(observed_y)[0]
        
        #Different layer models for each type of variable
        if types_list[i]['type'] == 'real':
            params = theta_real(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'pos':
            params = theta_pos(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'count':
            params = theta_count(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'cat':
            params = theta_cat(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'ordinal':
            params = theta_ordinal(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        theta.append(params)
            
    return theta

def theta_estimation_from_ys(samples_y, samples_s, types_list, miss_list, batch_size, reuse):
    
    theta = []
    
    #Independet yd -> Compute p(xd|yd)
    for i,d in enumerate(samples_y):
        
        #Partition the data in missing data (0) and observed data (1)
        missing_y, observed_y = tf.dynamic_partition(d, miss_list[:,i], num_partitions=2)
        missing_s, observed_s = tf.dynamic_partition(samples_s, miss_list[:,i], num_partitions=2)
        condition_indices = tf.dynamic_partition(tf.range(tf.shape(d)[0]), miss_list[:,i], num_partitions=2)
        nObs = tf.shape(observed_y)[0]
        
        #Different layer models for each type of variable
        if types_list[i]['type'] == 'real':
#            params = theta_real(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            params = theta_real_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'pos':
#            params = theta_pos(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            params = theta_pos_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'count':
#            params = theta_count(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            params = theta_count_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse)
        
        elif types_list[i]['type'] == 'cat':
#            params = theta_cat(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            params = theta_cat_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        elif types_list[i]['type'] == 'ordinal':
#            params = theta_ordinal(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse)
            params = theta_ordinal_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse)
            
        theta.append(params)
            
    return theta

def theta_real(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Mean layer
    h2_mean = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=True)

    #Sigma Layer
    h2_sigma = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2_sigma' + str(i), reuse=reuse, bias=True)
    
    return [h2_mean, h2_sigma]

def theta_real_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Mean layer    
    h2_mean = observed_data_layer(tf.concat([observed_y,observed_s],1), tf.concat([missing_y,missing_s],1), condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=False)
    
    #Sigma Layer    
    h2_sigma = observed_data_layer(observed_s, missing_s, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2_sigma' + str(i), reuse=reuse, bias=False)
    
    return [h2_mean, h2_sigma]


def theta_pos(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Mean layer
    h2_mean = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=True)
    
    #Sigma Layer
    h2_sigma = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2_sigma' + str(i), reuse=reuse, bias=True)
    
    return [h2_mean, h2_sigma]

def theta_pos_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Mean layer
    h2_mean = observed_data_layer(tf.concat([observed_y,observed_s],1), tf.concat([missing_y,missing_s],1), condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=False)
    
    #Sigma Layer
    h2_sigma = observed_data_layer(observed_s, missing_s, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2_sigma' + str(i), reuse=reuse, bias=False)
    
    return [h2_mean, h2_sigma]

def theta_count(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Lambda Layer
    h2_lambda = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=True)

    return h2_lambda

def theta_count_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Lambda Layer    
    h2_lambda = observed_data_layer(tf.concat([observed_y,observed_s],1), tf.concat([missing_y,missing_s],1), condition_indices, output_dim=types_list[i]['dim'], name='layer_h2' + str(i), reuse=reuse, bias=False)
    
    return h2_lambda

def theta_cat(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Log pi layer, with zeros in the first value to avoid the identificability problem
    h2_log_pi_partial = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=int(types_list[i]['dim'])-1, name='layer_h2' + str(i), reuse=reuse, bias=True)
    
    h2_log_pi = tf.concat([tf.zeros([batch_size,1]), h2_log_pi_partial],1)
    
    return h2_log_pi

def theta_cat_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Log pi layer, with zeros in the first value to avoid the identificability problem    
    h2_log_pi_partial = observed_data_layer(tf.concat([observed_y,observed_s],1), tf.concat([missing_y,missing_s],1), condition_indices, output_dim=int(types_list[i]['dim'])-1, name='layer_h2' + str(i), reuse=reuse, bias=False)
    
    h2_log_pi = tf.concat([tf.zeros([batch_size,1]), h2_log_pi_partial],1)
    
    return h2_log_pi

def theta_ordinal(observed_y, missing_y, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Theta layer, Dimension of ordinal - 1
    h2_theta = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=int(types_list[i]['dim'])-1, name='layer_h2' + str(i), reuse=reuse, bias=True)
    
    #Mean layer, a single value
    h2_mean = observed_data_layer(observed_y, missing_y, condition_indices, output_dim=1, name='layer_h2_sigma' + str(i), reuse=reuse, bias=True)
    
    return [h2_theta, h2_mean]

def theta_ordinal_s(observed_y, missing_y, observed_s, missing_s, condition_indices, types_list, nObs, batch_size, i, reuse):
    
    #Theta layer, Dimension of ordinal - 1
    h2_theta = observed_data_layer(observed_s, missing_s, condition_indices, output_dim=int(types_list[i]['dim'])-1, name='layer_h2' + str(i), reuse=reuse, bias=False)
    
    #Mean layer, a single value
    h2_mean = observed_data_layer(tf.concat([observed_y,observed_s],1), tf.concat([missing_y,missing_s],1), condition_indices, output_dim=1, name='layer_h2_sigma' + str(i), reuse=reuse, bias=False)
    
    return [h2_theta, h2_mean]

def observed_data_layer(observed_data, missing_data, condition_indices, output_dim, name, reuse, bias):
    
    #Train a layer with the observed data and reuse it for the missing data
    obs_output = tf.layers.dense(inputs=observed_data, units=output_dim, activation=None,
                         kernel_initializer=tf.random_normal_initializer(stddev=0.05),name=name,reuse=reuse,trainable=True,use_bias=bias)
    miss_output = tf.layers.dense(inputs=missing_data, units=output_dim, activation=None,
                     kernel_initializer=tf.random_normal_initializer(stddev=0.05),name=name,reuse=True,trainable=False,use_bias=bias)
    #Join back the data
    output = tf.dynamic_stitch(condition_indices, [miss_output,obs_output])
    
    return output


def loglik_evaluation(batch_data_list, types_list, miss_list, theta, tau2, normalization_params, reuse):
    
    log_p_x = []
    log_p_x_missing = []
    samples_x = []
    params_x = []
    
    #Independet yd -> Compute log(p(xd|yd))
    for i,d in enumerate(batch_data_list):
        
        # Select the likelihood for the types of variables
        loglik_function = getattr(loglik_models_missing_normalize, 'loglik_' + types_list[i]['type'])
        
        out = loglik_function([d,miss_list[:,i]], types_list[i], theta[i], normalization_params[i], tau2,
                                  kernel_initializer=tf.random_normal_initializer(stddev=0.05), name='layer_1_mean_dec_x' + str(i), reuse=reuse)
            
        log_p_x.append(out['log_p_x'])
        log_p_x_missing.append(out['log_p_x_missing']) #Test-loglik element
        samples_x.append(out['samples'])
        params_x.append(out['params'])
        
    return log_p_x, log_p_x_missing, samples_x, params_x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 28 10:56:13 2017

@author: anazabal, olmosUC3M, ivaleraM

List of loglikelihoods for the types of variables considered in this paper.
Basically, we create the different layers needed in the decoder and during the
generation of new samples

The variable reuse indicates the mode of this functions
- reuse = None -> Decoder implementation
- reuse = True -> Samples generator implementation

"""

import tensorflow as tf
import numpy as np

def loglik_real(batch_data,list_type,theta,normalization_params,tau2,kernel_initializer,name,reuse):
    
    output=dict()
    epsilon = tf.constant(1e-3, dtype=tf.float32)
    
    #Data outputs
    data, missing_mask = batch_data
    missing_mask = tf.cast(missing_mask,tf.float32)
    
    data_mean, data_var = normalization_params
    data_var = tf.clip_by_value(data_var, epsilon, np.inf)
    
    est_mean, est_var = theta
    est_var = tf.clip_by_value(tf.nn.softplus(est_var), epsilon, 1e20) #Must be positive
    
    # Affine transformation of the parameters
    est_mean = tf.sqrt(data_var)*est_mean + data_mean
    est_var = data_var*est_var
#    est_var = 0.05*tf.ones_like(est_var)
    
    #Compute loglik
    log_p_x = -0.5 * tf.reduce_sum(tf.squared_difference(data,est_mean)/est_var,1) - int(list_type['dim'])*0.5*tf.log(2*np.pi) - 0.5*tf.reduce_sum(tf.log(est_var),1)
#    log_p_x = -0.5 * tf.reduce_sum(tf.squared_difference(data,est_mean),1)
    
    #Outputs
    output['log_p_x'] = tf.multiply(log_p_x, missing_mask)
    output['log_p_x_missing'] = tf.multiply(log_p_x, 1.0-missing_mask)
    output['params'] = [est_mean, est_var]
    output['samples'] = tf.contrib.distributions.Normal(est_mean,tf.sqrt(est_var)).sample()
        
    return output


def loglik_pos(batch_data,list_type,theta,normalization_params,tau2,kernel_initializer,name,reuse):
    
    #Log-normal distribution
    output = dict()
    epsilon = tf.constant(1e-3, dtype=tf.float32)
    
    #Data outputs
    data_mean_log, data_var_log = normalization_params
    data_var_log = tf.clip_by_value(data_var_log, epsilon, np.inf)
    
    data, missing_mask = batch_data
    data_log = tf.log(1.0 + data)
    missing_mask = tf.cast(missing_mask,tf.float32)
    
    est_mean, est_var = theta
    est_var = tf.clip_by_value(tf.nn.softplus(est_var), epsilon, 1.0)
    
    # Affine transformation of the parameters
    est_mean = tf.sqrt(data_var_log)*est_mean + data_mean_log
    est_var = data_var_log*est_var
    

    #Compute loglik
    log_p_x = -0.5 * tf.reduce_sum(tf.squared_difference(data_log,est_mean)/est_var,1) \
        - 0.5*tf.reduce_sum(tf.log(2*np.pi*est_var),1) - tf.reduce_sum(data_log,1)
    
    output['log_p_x'] = tf.multiply(log_p_x, missing_mask)
    output['log_p_x_missing'] = tf.multiply(log_p_x, 1.0-missing_mask)
    output['params'] = [est_mean, est_var]
    output['samples'] = tf.clip_by_value(tf.exp(tf.contrib.distributions.Normal(est_mean,tf.sqrt(est_var)).sample()) - 1.0,0,1e20)
        
    return output

def loglik_cat(batch_data,list_type,theta,normalization_params,tau2,kernel_initializer,name,reuse):
    
    output=dict()
    
    #Data outputs
    data, missing_mask = batch_data
    missing_mask = tf.cast(missing_mask,tf.float32)
    
    log_pi = theta
    
    #Compute loglik
    log_p_x = -tf.nn.softmax_cross_entropy_with_logits(logits=log_pi,labels=data)
    
    output['log_p_x'] = tf.multiply(log_p_x, missing_mask)
    output['log_p_x_missing'] = tf.multiply(log_p_x, 1.0-missing_mask)
    output['params'] = log_pi
    output['samples'] = tf.one_hot(tf.contrib.distributions.Categorical(probs=tf.nn.softmax(log_pi)).sample(),depth=int(list_type['dim']))
    
    return output
    
def loglik_ordinal(batch_data,list_type,theta,normalization_params,tau2,kernel_initializer,name,reuse):
    
    output=dict()
    epsilon = tf.constant(1e-6, dtype=tf.float32)
    
    #Data outputs
    data, missing_mask = batch_data
    missing_mask = tf.cast(missing_mask,tf.float32)
    batch_size = tf.shape(data)[0]
    
    #We need to force that the outputs of the network increase with the categories
    partition_param, mean_param = theta
    mean_value = tf.reshape(mean_param,[-1,1])
    theta_values = tf.cumsum(tf.clip_by_value(tf.nn.softplus(partition_param), epsilon, 1e20),1)
    sigmoid_est_mean = tf.nn.sigmoid(theta_values - mean_value)
    mean_probs = tf.concat([sigmoid_est_mean,tf.ones([batch_size,1],tf.float32)],1) - tf.concat([tf.zeros([batch_size,1],tf.float32),sigmoid_est_mean],1)
    
    mean_probs = tf.clip_by_value(mean_probs,epsilon,1.0)
    
    #Code needed to compute samples from an ordinal distribution
    true_values = tf.one_hot(tf.reduce_sum(tf.cast(data,tf.int32),1)-1,int(list_type['dim']))
    
    #Compute loglik
    log_p_x = -tf.nn.softmax_cross_entropy_with_logits(logits=tf.log(mean_probs),labels=true_values)
    
    output['log_p_x'] = tf.multiply(log_p_x,missing_mask)
    output['log_p_x_missing'] = tf.multiply(log_p_x,1.0-missing_mask)
    output['params'] = mean_probs
    output['samples'] = tf.sequence_mask(1+tf.contrib.distributions.Categorical(logits=tf.log(tf.clip_by_value(mean_probs,epsilon,1e20))).sample(), int(list_type['dim']),dtype=tf.float32)
    
    return output
    
def loglik_count(batch_data,list_type,theta,normalization_params,tau2,kernel_initializer,name,reuse):
    
    output=dict()
    epsilon = tf.constant(1e-6, dtype=tf.float32)
    
    #Data outputs
    data, missing_mask = batch_data
    missing_mask = tf.cast(missing_mask,tf.float32)
    
    est_lambda = theta
    est_lambda = tf.clip_by_value(tf.nn.softplus(est_lambda),epsilon,1e20)
    
    log_p_x = -tf.reduce_sum(tf.nn.log_poisson_loss(targets=data,log_input=tf.log(est_lambda),compute_full_loss=True),1)
    
    output['log_p_x'] = tf.multiply(log_p_x, missing_mask)
    output['log_p_x_missing'] = tf.multiply(log_p_x, 1.0-missing_mask)
    output['params'] = est_lambda
    output['samples'] = tf.contrib.distributions.Poisson(est_lambda).sample()
        
    return output


import setuptools

with open("README.md", "r") as readme_file:
    readme = readme_file.read()

requirements = ['tensorflow>=1.13.0,<2',
                'pandas',
                'sklearn',
                    ]
    
setuptools.setup(
    name='hivae',
    version='0.19',
    url='https://github.com/gkoutos-group/hivae/',
    license='MIT',
    author='Andreas Karwath',
    author_email='a.karwath@bham.ac.uk',
    description='HIVAE (Handling incomplete heterogeneous data using VAEs. - by Nazabal, et al., DOI: 10.1016/j.patcog.2020.107501, 2020) - Extenstion of implementations as easy to use Python library',
    packages=setuptools.find_packages(exclude=['examples']),
    install_requires=requirements,
    long_description=readme,
    long_description_content_type="text/markdown",
    zip_safe=False
    )


